unit PollinLCD;

//gendert: 2002-03-12

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, PollinLCDBasics, ExFont;

type
  TDisplayMode = (dmOR, dmXOR, dmAND);
  TRefreshTextEvent = procedure (Owner: TObject; TextData: array of Byte) of object;
  TRefreshGraphEvent = procedure (Owner: TObject; GraphData: array of Byte) of object;
  TRefreshFontEvent = procedure (Owner: TObject; FontData: array of Byte) of object;

  TTextData = array[1 .. TEXT_ROWS * TEXT_COLUMS] of Byte;
  TGraphData = array[1 .. GRAPH_ROWS * GRAPH_COLUMS] of Byte;
  TFontData = array[0..7*255] of Byte;

  TPollinLCD = class(TComponent)
  private
    _LPTDriverstarted: Boolean;
    _GraficEnabled: Boolean;
    _TextEnabled: Boolean;
    _Active: Boolean;
    _PrinterPort: Integer;
    _BlinkingCursor: Boolean;
    _Bidirectional: Boolean;
    _CursorEnabled: Boolean;
    _DisplayMode: Integer;
    _CursorPosY: Integer;
    _CursorPosX: Integer;
    _CursorHeight: Integer;
    _DOSFont: Boolean;
    _ExternalFont: Boolean;

    _Font: TFont;

    _TextData: TTextData;
    _OldTextData:  TTextData;
    _GraphData: TGraphData;
    _OldGraphData:  TGraphData;
    _FontData: TFontData;


    _OnTextRefresh: TRefreshTextEvent;
    _OnGraphRefresh: TRefreshGraphEvent;
    _OnFontRefresh: TRefreshFontEvent;

    procedure SetActive(const Value: Boolean);
    procedure SetTextEnabled(const Value: Boolean);
    procedure SetGraficEnabled(const Value: Boolean);
    procedure SetDisplayMode(const Value: Integer);
    procedure SetExternalFont(const Value: Boolean);
    procedure SetCursorEnabled(const Value: Boolean);
    procedure SetCursorBlink(const Value: Boolean);
    procedure SetCursorPosX(const Value: Integer);
    procedure SetCursorPosY(const Value: Integer);
    procedure SetCursorHeight(const Value: Integer);
    procedure SetFont(const Value: TFont);
    function GetText: String;
    procedure SetText(const Value: String);
    function GetLine(Index: Integer): String;
    procedure SetLine(Index: Integer; const Value: String);
    procedure SetDOSFont(const Value: Boolean);
  protected
  public
    IsUpdating: Boolean;
    c, cc, dc: Integer;
    constructor Create(AOwner: TComponent);   override;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Clear;
    procedure ClearText;
    procedure ClearGraphic;
    procedure Refresh;
    procedure RefreshText;
    procedure RefreshGraphic;
    procedure DrawText(x, y: Integer; Str: String);
    procedure DrawBmp(dx, dy: Integer; Bmp: TBitmap);
    procedure DrawLine(SPosX,SPosY,EPosX,EPosY: Integer; IsBlack: Boolean);

    property Lines[Index: Integer]: String read GetLine write SetLine;
  published
    property Active: Boolean read _Active write SetActive;                      //Display Aktiv
    property DisplayMode: Integer read _DisplayMode write SetDisplayMode;       //OR XOR AND
    property DOSFont: Boolean read _DOSFont write SetDOSFont;
    property ExternalFont: Boolean read _ExternalFont write SetExternalFont;
    property EnabledText: Boolean read _TextEnabled write SetTextEnabled;       //Text Anzeigen
    property EnabledGraphic: Boolean read _GraficEnabled write SetGraficEnabled; //Grafik anzeigen
    property EnabledCursor: Boolean read _CursorEnabled write SetCursorEnabled; //Cursor anzeigen
    property CursorBlink: Boolean read _BlinkingCursor write SetCursorBlink;    //Cursor blinken

    property CursorPosX: Integer read _CursorPosX write SetCursorPosX;          //Cursor Position
    property CursorPosY: Integer read _CursorPosY write SetCursorPosY;
    property CursorHeight: Integer read _CursorHeight write SetCursorHeight;    //Cursor Grsse

    property PrinterPort: Integer read _PrinterPort write _PrinterPort;         //DruckerPort
    property Bidirectional: Boolean read _Bidirectional write _Bidirectional;   //Bidirektionaler Betrieb
    property Font: TFont read _Font write SetFont;
    property Text: String read GetText write SetText;
    property OnTextRefresh: TRefreshTextEvent read _OnTextRefresh write _OnTextRefresh;
    property OnGraphRefresh: TRefreshGraphEvent read _OnGraphRefresh write _OnGraphRefresh;
    property OnFontRefresh: TRefreshFontEvent read _OnFontRefresh write _OnFontRefresh;
   end;

procedure Register;

implementation

uses PollinLCDPreview;

procedure Register;
begin
  RegisterComponents('LCD', [TPollinLCD, TPollinLCDPreview]);
end;

{ TPollinLCD }

constructor TPollinLCD.Create(AOwner: TComponent);
begin
  inherited;

  Fillchar(_TextData, sizeOf(_TextData), #0);
  Fillchar(_GraphData, sizeOf(_GraphData), #0);
  Fillchar(_OldTextData, sizeOf(_OldTextData), #255); //_OldTextData entspricht nach dem Initialisieren nicht dem DisplayText
  Fillchar(_OldGraphData, sizeOf(_OldGraphData), #255); //_OldTextData entspricht nach dem Initialisieren nicht dem DisplayText

  _Active := False;
  _LPTDriverstarted := StartLPTDriver;
  _PrinterPort    := $378;        //standard DruckerPort LPT1
  _Bidirectional  := False;      //Bidirektonial ECP
  _CursorPosY := 1;
  _CursorPosX := 1;
  _CursorHeight := 1;
  InitLCD(_PrinterPort, _Bidirectional); //DisplayMode CursorOff, BlinkOff, TextOff, GraficOff
  _Font := TFont.Create;
  IsUpdating := False;
  c := 0;
  cc := 0;
  dc := 0;
end;

destructor TPollinLCD.Destroy;
begin
  inherited;
end;

//------------------------------------------------------------------------------
//                                  private
//------------------------------------------------------------------------------

procedure TPollinLCD.SetActive(const Value: Boolean);
begin
  if (_LPTDriverstarted) and (Value <> _Active) then
  begin
        if Value then
              InitLCD(_PrinterPort, _Bidirectional);
        _Active := Value;
  end;
end;

procedure TPollinLCD.SetTextEnabled(const Value: Boolean);
begin
  if not _Active then Exit;
  _TextEnabled := Value;
  SetCursorEnabled(_CursorEnabled);
end;

procedure TPollinLCD.SetGraficEnabled(const Value: Boolean);
begin
  if not _Active then Exit;
  _GraficEnabled := Value;
  SetCursorEnabled(_CursorEnabled);
end;

procedure TPollinLCD.SetDisplayMode(const Value: Integer);
begin
  if not _Active then Exit;
  // Mode gendert
  _DisplayMode := Value;
  if _ExternalFont then
  begin
       case _DisplayMode of
            dm_or  : WriteCtrl(_PrinterPort, _Bidirectional, $80 + 8); (* OR Mode *)
            dm_xor : WriteCtrl(_PrinterPort, _Bidirectional, $81 + 8);
            dm_and : WriteCtrl(_PrinterPort, _Bidirectional, $83 + 8);
       end;
  end
  else
  begin
       case _DisplayMode of
            dm_or  : WriteCtrl(_PrinterPort, _Bidirectional, $80); (* OR Mode *)
            dm_xor : WriteCtrl(_PrinterPort, _Bidirectional, $81);
            dm_and : WriteCtrl(_PrinterPort, _Bidirectional, $83);
       end;
  end;
end;

procedure TPollinLCD.SetExternalFont(const Value: Boolean);
begin
  if not _Active then Exit;
  _ExternalFont := Value;
  SetDisplayMode(_DisplayMode);
  Fillchar(_OldTextData, sizeOf(_OldTextData), #255); //_OldTextData entspricht nach dem Initialisieren nicht dem DisplayText
  if not IsUpdating then RefreshText;
end;

procedure TPollinLCD.SetCursorEnabled(const Value: Boolean);
var b: Byte;
begin
  if not _Active then Exit;
  // Darstellung (Text,Grafik,Cursor) gendert, neu bertragen
  _CursorEnabled := Value;
  b := $90;
  if _BlinkingCursor then b := b or 1;
  if _CursorEnabled  then b := b or 2;
  if _GraficEnabled  then b := b or 8;
  if _TextEnabled    then b := b or 4;
  WriteCtrl(_PrinterPort, _Bidirectional, b);
end;

procedure TPollinLCD.SetCursorBlink(const Value: Boolean);
begin
  if not _Active then Exit;
  _BlinkingCursor := Value;
  SetCursorEnabled(_CursorEnabled);
end;

procedure TPollinLCD.SetCursorPosX(const Value: Integer);
begin
  if not _Active then Exit;
  _CursorPosX := IntMinMax(Value, 1, TEXT_COLUMS);

  WriteData(_PrinterPort, _Bidirectional, _CursorPosX - 1);
  WriteData(_PrinterPort, _Bidirectional, _CursorPosY - 1);
  WriteCtrl(_PrinterPort, _Bidirectional, $21); //Command: REGISTERSSETTING Code:$21 --> 00100001
end;

procedure TPollinLCD.SetCursorPosY(const Value: Integer);
begin
  if not _Active then Exit;
  _CursorPosY := IntMinMax(Value, 1, TEXT_ROWS);

  WriteData(_PrinterPort, _Bidirectional, _CursorPosX - 1);
  WriteData(_PrinterPort, _Bidirectional, _CursorPosY - 1);
  WriteCtrl(_PrinterPort, _Bidirectional, $21); //Command: REGISTERSSETTING Code:$21 --> 00100001
end;

procedure TPollinLCD.SetCursorHeight(const Value: Integer);
begin
  if not _Active then Exit;
  _CursorHeight := IntMinMax(Value, 1, 8);
  WriteCtrl(_PrinterPort, _Bidirectional, $A0 + _CursorHeight - 1);
end;

procedure TPollinLCD.SetFont(const Value: TFont);
var tmpbit: TBitmap;
    zae,x,y,d,xo,yo: integer;
begin
  if not _Active then Exit;
  _Font := Value;
  Fillchar(_FontData, SizeOf(_FontData), #0);

  // Schriftart bertragen
  tmpbit := TBitmap.create;
  tmpbit.width:=50;
  tmpbit.height:=50;
  tmpbit.canvas.font:=_Font;

  // Schreibadresse auf CG-RAM
  SetAddressPointer(_PrinterPort, _Bidirectional, CGRAM_HOME_ADDRESS);

  WriteCtrl(_PrinterPort, _Bidirectional, $b0);   //auto data write

  with tmpbit.canvas do
  begin
       // y-Offset
       y := TextHeight('g');
       yo := (y-8) div 2; if yo < 0 then yo := 0;

       //#0 immer als "Leerzeichen"
       WriteData(_PrinterPort, _Bidirectional, 0);
       for x := 0 to 7 do _FontData[x] := 0;

       // Zeichen ins Bitmap malen und ans Display bertragen
       for zae := 1 to 255 do
       begin
            // x-Offset
            x := TextWidth(chr(zae));
            xo := (x-6) div 2; if xo<0 then xo:=0;

            // Zeichenflche lschen
            pen.color:=clWhite;
            rectangle(0,0,49,49);
            pen.color:=clBlack;

            // Zeichen ausgeben
            textout(0,0,chr(zae));

            // Pixelweise abtasten und Fontdaten schreiben
            for y := 0 to 7 do
            begin
                 d:=0;
                 for x:=0 to 5 do
                       if pixels[x+xo,y+yo]<>clWhite then d:=d or (1 shl (5-x));
                 WriteData(_PrinterPort, _Bidirectional, d);
                 _FontData[zae*8+y] := d;
            end;
       end;
  end;
  WriteCtrl(_PrinterPort, _Bidirectional, $b2);  //auto read/write reset

  tmpbit.free;
  if not IsUpdating then RefreshText;
  if Assigned(_OnFontRefresh) then _OnFontRefresh(Self, _FontData);
end;

function TPollinLCD.GetText: String;
var i : Integer;
begin
  Result := '';
  for i := 1 to High(_TextData) do
  begin
        if _TextData[i] = 0 then Break;
        Result := Result + Chr(_TextData[i]);
  end;
end;

procedure TPollinLCD.SetText(const Value: String);
var i : Integer;
begin
  if not _Active then Exit;
  FillChar(_TextData, SizeOf(_TextData), #0);
  if Value <> '' then
     for i := 1 to IntMinMax(Length(Value), 1, TEXT_ROWS*TEXT_COLUMS) do
        _TextData[i] := Ord(Value[i]);
  if not IsUpdating then RefreshText;
end;

function TPollinLCD.GetLine(Index: Integer): String;
var i : Integer;
begin
  Result := '';
  for i := (Index-1)*TEXT_COLUMS +1 to Index*TEXT_COLUMS do
  begin
        if _TextData[i] = 0 then Break;
        Result := Result + Chr(_TextData[i]);
  end;
end;

procedure TPollinLCD.SetLine(Index: Integer; const Value: String);
var i: Integer;
begin
  if not _Active then Exit;
  if Value <> '' then
     for i := 1 to IntMinMax(Length(Value), 1, TEXT_COLUMS) do
         _TextData[(Index-1)*TEXT_COLUMS + i] := Ord(Value[i]);
  if not IsUpdating then RefreshText;
end;

//------------------------------------------------------------------------------
//                                  public
//------------------------------------------------------------------------------

procedure TPollinLCD.BeginUpdate;
begin
  IsUpdating := True;
end;

procedure TPollinLCD.EndUpdate;
begin
  IsUpdating := False;
  Refresh;
end;

procedure TPollinLCD.Clear;
begin
  if _GraficEnabled then ClearGraphic;  //Zeit sparen durch seperates Lschen
  if _TextEnabled then ClearText;
end;

procedure TPollinLCD.ClearText;
begin
  if not _Active then Exit;
  Fillchar(_TextData, sizeOf(_TextData), #0);
  if not IsUpdating then RefreshText;
end;

procedure TPollinLCD.ClearGraphic;
begin
  if not _Active then Exit;
  Fillchar(_GraphData, sizeOf(_GraphData), #0);
  if not IsUpdating then RefreshGraphic;
end;

procedure TPollinLCD.Refresh;
begin
  if _TextEnabled then RefreshText;
  if _GraficEnabled then RefreshGraphic;
end;

procedure TPollinLCD.RefreshText;
var
  Changed: array[1..TEXT_ROWS*TEXT_COLUMS] of Boolean;

  function WriteUnChanged(Index: Integer): Boolean;
  var k: Integer;
  begin
    k := 1;                    //Zhlen wie viele Zeichen folgen, die sich nicht gendert haben
    while not Changed[Index+k] do
    begin
          if Index+k > High(Changed) then Break;
          if k >  4 then Break;
          inc(k);
    end;                      //Wenn sich nach Index weniger als 4 nicht genderte Zeichen befinden
    Result := (k < 4);        //sollen diese mitgeschrieben werden, um Schreibaktionen zu sparen
  end;
var i: Integer;
begin
  if not _Active then Exit;
  if _LPTDriverstarted then
  begin
        //"cleveres", weil dadurch schnelleres Updaten der Anzeige
        //es wird nur das neu ins Display geschrieben was sich gendert hat
        for i := 1 to High(Changed) do
            Changed[i] := (_OldTextData[i] <> _TextData[i]); //Hat sich etwas gendert?
        i := 1;
        while i <= High(_TextData) do
        begin
              if (Changed[i]) and (i <= High(_TextData)) then
              begin
                    SetAddressPointer(_PrinterPort, _Bidirectional, TEXT_HOME_ADDRESS + i - 1);
                    WriteCtrl(_PrinterPort, _Bidirectional, $b0);   //auto data write
                    Inc(cc, 2);  //SetAddr.P. -> + 2*Data + 1*Ctrl
                    Inc(dc, 2);  //WriteCtrl  -> + 1*Ctrl
                    while (Changed[i] or WriteUnChanged(i)) and (i <= High(_TextData)) do
                    begin
                          if _ExternalFont then
                                WriteData(_PrinterPort, _Bidirectional, _TextData[i])
                          else
                                WriteData(_PrinterPort, _Bidirectional, ANSI2DisplayText(_TextData[i]));
                          if Changed[i] then Inc(c);
                          Inc(i);
                          Inc(dc); //WriteData -> + 1*Data
                    end;
                    WriteCtrl(_PrinterPort, _Bidirectional, $b2);  //auto read/write reset
                    Inc(cc);       //WriteCtrl -> + 1*Ctrl
              end
              else
                    Inc(i);
        end;
        for i := 1 to High(_TextData) do _OldTextData[i] := _TextData[i];
  end;
 if Assigned(_OnTextRefresh) then _OnTextRefresh(Self, _TextData);
end;

procedure TPollinLCD.RefreshGraphic;
var
  Changed: array[1..GRAPH_ROWS*GRAPH_COLUMS] of Boolean;

  function WriteUnChanged(Index: Integer): Boolean;
  var k: Integer;
  begin
    k := 1;                    //Zhlen wie viele Zeichen folgen, die sich nicht gendert haben
    while not Changed[Index+k] do
    begin
          if Index+k > High(Changed) then Break;
          if k >  4 then Break;
          inc(k);
    end;                      //Wenn sich nach Index weniger als 4 nicht genderte Zeichen befinden
    Result := (k < 4);        //sollen diese mitgeschrieben werden, um Schreibaktionen zu sparen
  end;

var i: Integer;
begin
  if not _Active then Exit;
  if _LPTDriverstarted then
  begin
        for i := 1 to High(Changed) do
            Changed[i] := (_OldGraphData[i] <> _GraphData[i]); //Hat sich etwas gendert?
        i := 1;
        while i <= High(_GraphData) do
        begin
              if (Changed[i]) and (i <= High(_GraphData)) then
              begin
                    SetAddressPointer(_PrinterPort, _Bidirectional, GRAPH_HOME_ADDRESS + i - 1);
                    WriteCtrl(_PrinterPort, _Bidirectional, $b0);   //auto data write
                    Inc(cc, 2);
                    Inc(dc, 2);
                    while (Changed[i] or WriteUnChanged(i)) and (i <= High(_GraphData)) do
                    begin
                          WriteData(_PrinterPort, _Bidirectional, _GraphData[i]);
                          if Changed[i] then Inc(c);
                          Inc(i);
                          Inc(dc);
                    end;
                    WriteCtrl(_PrinterPort, _Bidirectional, $b2);  //auto read/write reset
                    Inc(cc);
              end
              else
                    Inc(i);
        end;
        for i := 1 to High(_GraphData) do _OldGraphData[i] := _GraphData[i];
  end;
 if Assigned(_OnGraphRefresh) then _OnGraphRefresh(Self, _GraphData);
end;

procedure TPollinLCD.DrawText(x, y: Integer; Str: String);
begin

end;

procedure TPollinLCD.DrawBmp(dx, dy: Integer; Bmp: TBitmap);
var
  x,y, z,i : Integer;
  p: Integer;
  pp: Single;
  TmpBmp: TBitmap;
begin
  if not _Active then Exit;
  tmpBmp := TBitmap.Create;
  tmpBmp.Height := bmp.Height + dy;
  tmpBmp.Width := bmp.Width + dx;

  tmpBmp.Canvas.CopyRect(Rect(dx, dy, dx+bmp.Width, dy+bmp.Height), Bmp.Canvas, Rect(0, 0, bmp.Width, bmp.Height));

  for y := 0 to 63 do
    for x:=0 to (240 div 6)-1 do begin
      i:=0;
      for z:=0 to 5 do begin
        // Pixel schreiben; die Random-Funktion dient zum Dithern
        // (Streuraster) von Farbgrafiken
        p:=tmpBmp.canvas.Pixels[x*6+z,y];
        pp:=0.6*(p AND $FF)+0.30*((p shr 8) AND $FF)+0.1*((p shr 16) AND $FF);
        i:=i shl 1;
        if pp<=random(255) then i:=i or 1;
      end;
      _GraphData[y*GRAPH_COLUMS+x+1] := _GraphData[y*GRAPH_COLUMS+x+1] or i;
    end;
  tmpBmp.Free;
  if not IsUpdating then RefreshGraphic;
end;

procedure TPollinLCD.DrawLine(SPosX,SPosY,EPosX,EPosY: Integer; IsBlack: Boolean);

          procedure SetPixel(x, y: Integer; IsBlack: Boolean);
          var DataPos: Integer; b, newb: Byte;
          begin
               DataPos := (y-1)*GRAPH_COLUMS+ ((x-1) div 6)+1;
               b := _GRAPHDATA[DataPos];
               newb := 1 shl (5-((x-1) mod 6));
               if IsBlack then
                   _GRAPHDATA[DataPos] := b or newb
               else
                   _GRAPHDATA[DataPos] := b and (not newb);
          end;
{vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv}
{ Folgende Zeilen sind NICHT von mir und ich weis nicht mehr von wem ich       }
{ diese Hilfe bekommen habe. Irgendwo ausm Internet.                           }
          procedure SwapInt(var PosX,PosY:Integer);
          var Dummy: Integer;
          begin
               Dummy := PosY; PosY := PosX; PosX := Dummy;
          end;

Var Puffer,X1,Y1,X2,Y2,IncrX1,IncrY1,IncrX2,IncrY2:Integer;
Begin
  If (Abs(EPosX-SPosX)<Abs(EPosY-SPosY)) Then
  Begin
       If (SPosY>EPosY) Then
       Begin
             SwapInt(SPosX,EPosX);
             SwapInt(SPosY,EPosY);
       End;
       If (EPosX>SPosX) Then IncrX2:=1 Else IncrX2:=-1;
       Y2:=EPosY-SPosY;
       X2:=Abs(EPosX-SPosX);
       Puffer:=2*X2-Y2;
       IncrX1:=2*(X2-Y2);
       IncrY1:=2*X2;
       X1:=SPosX;
       Y1:=SPosY;
       SetPixel(X1,Y1, IsBlack);
       For Y1:=SPosY+1 To EPosY Do
       Begin
             If (Puffer>=0) Then
             Begin
                   Inc(X1,IncrX2);
                   Inc(Puffer,IncrX1);
             End
             Else
                   Inc(Puffer,IncrY1);
             SetPixel(X1,Y1, IsBlack);
       End;
  End
  Else
  Begin
        If (SPosX>EPosX) Then
        Begin
              SwapInt(SPosX,EPosX);
              SwapInt(SPosY,EPosY);
        End;
        If (EPosY>SPosY) Then IncrY2:=1 Else IncrY2:=-1;
        X2:=EPosX-SPosX;
        Y2:=Abs(EPosY-SPosY);
        Puffer:=2*Y2-X2;
        IncrX1:=2*(Y2-X2);
        IncrY1:=2*Y2;
        X1:=SPosX;
        Y1:=SPosY;
        SetPixel(X1,Y1, IsBlack);
        For X1:=SPosX+1 To EPosX Do
        Begin
              If (Puffer>=0) Then
              Begin
                    Inc(Y1,IncrY2);
                    Inc(Puffer,IncrX1);
              End
              Else
                    Inc(Puffer,IncrY1);
              SetPixel(X1,Y1, IsBlack);
        End;
  End;
{^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^}
  if not IsUpdating then RefreshGraphic;
end;

procedure TPollinLCD.SetDOSFont(const Value: Boolean);
begin
  if not _Active then Exit;
  if Value then
  begin
      LoadExFont(printerport, Bidirectional);
      SetExternalFont(true);
  end
  else
      SetExternalFont(False);
  _DOSFont := Value;
end;


end.




